#include "Wire.h"

#include "I2Cdev.h"
#include "MPU6050.h"
#include "math.h"
#include <AccelStepper.h>
#include <MultiStepper.h>
#include <avr/wdt.h>
#include <SoftwareSerial.h>

#define leftMotorPWMPin   9
#define leftMotorDirPin   8
#define rightMotorPWMPin  11
#define rightMotorDirPin  10
#define RxD 5
#define TxD 6
#define pirPin 2
#define buzzerPin 13
#define Kp  15
#define Kd  0.1
#define Ki 0
#define sampleTime  0.005
#define targetAngle -0.5

MPU6050 mpu;
SoftwareSerial BTSerial(RxD, TxD);
// Initialize AccelStepper for two motors
AccelStepper stepper1(AccelStepper::DRIVER, leftMotorPWMPin, leftMotorDirPin);
AccelStepper stepper2(AccelStepper::DRIVER, rightMotorPWMPin, rightMotorDirPin);

int16_t accY, accZ, accX, gyroX;
volatile int motorPower, gyroRate;
volatile float accAngle, gyroAngle, currentAngle, prevAngle=0, error, prevError=0, errorSum=0;
volatile byte count=0;
volatile uint32_t prevTime = 0;
uint32_t currTime = 0;
char moveState = 'S';
volatile int flag = 0;
volatile int activeBuzzer = 0;
int distanceCm;


// Function to be called when the interrupt occurs
void handleInterrupt() {
  // Code to execute when the interrupt occurs
  if (activeBuzzer) {
    digitalWrite(buzzerPin, HIGH);
  }
  prevTime = millis();
  flag = 1;
}

void init_PID() {  
  // initialize Timer1
  cli();          // disable global interrupts
  TCCR1A = 0;     // set entire TCCR1A register to 0
  TCCR1B = 0;     // same for TCCR1B    
  // set compare match register to set sample time 5ms
  OCR1A = 9999;    
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS11 bit for prescaling by 8
  TCCR1B |= (1 << CS11);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();          // enable global interrupts
}

void setup() {
  Serial.begin(115200);
  //Serial.println("started init");
  // set the motor control and PWM pins to output mode
  pinMode(leftMotorPWMPin, OUTPUT);
  pinMode(leftMotorDirPin, OUTPUT);
  pinMode(rightMotorPWMPin, OUTPUT);
  pinMode(rightMotorDirPin, OUTPUT);
  pinMode(pirPin, INPUT);
  pinMode(buzzerPin, OUTPUT); 
  BTSerial.begin(9600);
  // initialize the MPU6050 and set offset values
  //Serial.println("started mpu");
  mpu.initialize();
  mpu.setXAccelOffset(594);
  mpu.setYAccelOffset(3774);
  mpu.setZAccelOffset(1398);
  mpu.setXGyroOffset(148);
  mpu.setYGyroOffset(86);
  mpu.setZGyroOffset(6);
  //Serial.println("finished mpu");
  stepper1.setMaxSpeed(3000);
  stepper1.setAcceleration(3000);
  stepper2.setMaxSpeed(3000);
  stepper2.setAcceleration(3000);

  attachInterrupt(digitalPinToInterrupt(pirPin), handleInterrupt, RISING);
  // initialize PID sampling loop
  init_PID();
  wdt_enable(WDTO_1S);
  //Serial.println("finished init");
}

volatile bool updateFlag = false; // Flag to signal when to update PID calculations

ISR(TIMER1_COMPA_vect) {
    updateFlag = true; // Set the flag to true to indicate data should be processed
}

void loop() {

    wdt_reset();

    if (BTSerial.available()) {
      char c =  BTSerial.read();
      if (c == 'A') {
        activeBuzzer = 1 - activeBuzzer;
      } else {
        moveState = c;
      }
      Serial.print(c);
    }
    
    currTime = millis();
    if (currTime - prevTime >= 1000) {
      digitalWrite(buzzerPin, LOW);
    }

    if (flag) {
      Serial.println("Motion");
      Serial.println(millis());
      BTSerial.write('a');
      flag = 0;
    }


    if (updateFlag) {
        // Reset the flag
        updateFlag = false;

        // Perform sensor reading and calculations
        readSensors();
        calculatePID();

        // Update motors
        stepper1.setSpeed(-motorPower);
        stepper2.setSpeed(motorPower);
        
        //Serial.println(motorPower);
    }

    stepper1.runSpeed();
    stepper2.runSpeed();
}

void readSensors() {
    //Serial.println("Reading sensor");
    accY = mpu.getAccelerationY();
    accZ = mpu.getAccelerationZ();
    accX = mpu.getAccelerationX();  
    gyroX = mpu.getRotationX();

    accAngle = atan2(accX, accZ) * RAD_TO_DEG;
    gyroRate = map(gyroX, -32768, 32767, -250, 250);
    gyroAngle = (float)gyroRate * sampleTime;
    //Serial.println("Finished reading sensors");
}

void calculatePID() {
    //Serial.println("Calculating");
    currentAngle = 0.9934 * (prevAngle + gyroAngle) + 0.0066 * (accAngle);
    error = currentAngle - targetAngle;
    errorSum += error;  
    errorSum = constrain(errorSum, -300, 300);
    motorPower = Kp * error + Ki * errorSum * sampleTime - Kd * (currentAngle - prevAngle) / sampleTime;

    motorPower = constrain(motorPower, -3000, 3000);
    prevAngle = currentAngle;
    //Serial.println("Finished calculating");
}